                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/


void Polygon::initialize(Polygon *&poly, unsigned int nVerts){
	poly = new Polygon;

	poly->nVertices = nVerts;
	poly->vertices = new Vertex[nVerts];
	poly->texID  = texID;
    poly->bumpID = bumpID;
	poly->s = s;
	poly->t = t;
    poly->scaleS = scaleS;
    poly->scaleT = scaleT;
	poly->origo = origo;
	poly->flags = flags;
}

void Polygon::finalize(){
	normal = cross(vertices[1] - vertices[0], vertices[nVertices - 1] - vertices[0]);
	normal.normalize();
	offset = -(normal * vertices[0]);
}

void Polygon::copyVertex(Polygon *&poly, unsigned int destIndex, unsigned int srcIndex){
	poly->vertices[destIndex] = vertices[srcIndex];
}

void Polygon::setInterpolatedVertex(Polygon *&poly, unsigned int destIndex, int srcIndex0, int srcIndex1, float k){
	poly->vertices[destIndex] = vertices[srcIndex0] + k * (vertices[srcIndex1] - vertices[srcIndex0]);
}

bool Polygon::linePassesThrough(const Vertex &v0, const Vertex &v1, Vertex &result){
	float k;
	Vertex p,g0,g1,g2;//,f;

	// Find interpolation constant
	k = normal * (v0 - v1);
	if (k == 0.0f) return false;
	k = (normal * v0 + offset) / k;

	// If not within the points, then it doesn't even hit the plane
	if (k <= 0 || k >= 1) return false;

	// Find the point it hits the polygons plane, p is the actual point, f is relative first vertex
	p = v0 + k * (v1 - v0);
	//f = p - vertices[0];





	g1 = vertices[1] - vertices[0];
	for (int i = 2; i < nVertices; i++){
		/*g0 = g1;
		g1 = vertices[i] - vertices[0];

		a = length(cross(g0,f)) + length(cross(g1,f)) + length(cross(g0 - g1, f - g1));
		area = length(cross(g0, g1));

		// There must be a better way than comparing areas ...
		// but it does at least work ...
		if (fabs(a - area) < 1){
			result = p;
			return true;
		}
		if (fabsf(a0 + a1 + a2 - 1) < 0.1f) return true;
		*/



		// Solve the P = a0 * P0 + a1 * P1 + a2 * P2 equation
		g0 = vertices[0];
		g1 = vertices[i - 1];
		g2 = vertices[i];


		float a0,a1,a2, M[9];
		float invDetM = 1.0f / (g0.x * (g1.y * g2.z - g1.z * g2.y) -
		                        g1.x * (g0.y * g2.z - g0.z * g2.y) + 
					            g2.x * (g0.y * g1.z - g0.z * g1.y));

		M[0] =  (g1.y * g2.z - g1.z * g2.y) * invDetM;
		M[1] = -(g1.x * g2.z - g1.z * g2.x) * invDetM;
		M[2] =  (g1.x * g2.y - g1.y * g2.x) * invDetM;
		
		M[3] = -(g0.y * g2.z - g0.z * g2.y) * invDetM;
		M[4] =  (g0.x * g2.z - g0.z * g2.x) * invDetM;
		M[5] = -(g0.x * g2.y - g0.y * g2.x) * invDetM;
		
		M[6] =  (g0.y * g1.z - g0.z * g1.y) * invDetM;
		M[7] = -(g0.x * g1.z - g0.z * g1.x) * invDetM;
		M[8] =  (g0.x * g1.y - g0.y * g1.x) * invDetM;

		a0 = M[0] * p.x + M[1] * p.y + M[2] * p.z;
		a1 = M[3] * p.x + M[4] * p.y + M[5] * p.z;
		a2 = M[6] * p.x + M[7] * p.y + M[8] * p.z;



		
/*		g0 = g1;
		g1 = vertices[i] - vertices[0];
		
		float a0,a1,a2;
		float invDetM, detM = g0.x * g1.y - g0.y * g1.x;
		if (detM != 0){
			invDetM = 1.0f / detM;
			a0 = (g1.y * f.x - g1.x * f.y) * invDetM;
			a1 = (g0.x * f.y - g0.y * f.x) * invDetM;
		} else {
			detM = g0.x * g1.z - g0.z * g1.x;
			invDetM = 1.0f / detM;
			if (detM != 0){
				a0 = (g1.z * f.x - g1.x * f.z) * invDetM;
				a1 = (g0.x * f.z - g0.z * f.x) * invDetM;
			} else {
				invDetM = 1.0f / (g0.y * g1.z - g0.z * g1.y);
				a0 = (g1.z * f.y - g1.y * f.z) * invDetM;
				a1 = (g0.y * f.z - g0.z * f.y) * invDetM;
			}
		}
		a2 = 1.0f - a0 - a1;
*/	


		//if (a0 >= -0.01f && a0 <= 1.01f && a1 >= -0.01 && a1 <= 1.01f && a2 >= -0.01f && a2 <= 1.01f) return true;
		if (a0 >= 0 && a0 <= 1 && a1 >= 0 && a1 <= 1 && a2 >= 0 && a2 <= 1){
			result = p;
			return true;
		}


	}

	return false;
}

bool Polygon::sphereIntersects(const Vertex &v, const float radius){
    // Note: This function may fail if the point is central to a huge polygon such that it doesn't intersect any of it's edges
    float d, distance;
    Vertex g0, g1, a, b;

    g1 = vertices[nVertices - 1];
    int i, len = nVertices;
    for (i = 0; i < len; i++){
        g0 = g1;
        g1 = vertices[i];

        b = g1 - g0;
        b.normalize();
        a = cross(normal, b);

        distance = fabsf((v - g0) * a);
        if (distance < radius){
            d = (v - g0) * b;
            if ((d > 0 && d / length(g1 - g0) < 1) || length(v - g0) < radius){
                return true;
            }
        }
    }
    return false;
}

RELATION Polygon::split(Polygon &poly, Polygon *&back, Polygon *&front){
	RELATION rel, prev, rel0, rel1, relprev0;
	int i, j, split0, split1, offset;
	int nBack, nFront;
	float k;


	nFront = nBack = 0;
	split0 = -1, split1 = -1;

	prev = vertexRelation(poly.vertices[poly.nVertices - 1]);

	for (i = 0; i < poly.nVertices; i++){
		rel = vertexRelation(poly.vertices[i]);

		if (rel == BACK)  nBack++;
		if (rel == FRONT) nFront++;

		if (rel != prev){
			if (prev != PLANAR){
				if (split0 < 0){
					split0 = i;
					rel0 = rel;
					relprev0 = prev;
				} else {
					split1 = i;
					rel1 = rel;
				}
			}
		}

		prev = rel;
	}

	if (nBack == 0){
		back = front = NULL;
		if (nFront == 0){
			return PLANAR;
		} else {
			return FRONT;
		}
	}

	if (nFront == 0){
		back = front = NULL;
		return BACK;
	}
	
	if (relprev0 == FRONT){
		int temp = nFront;
		nFront = nBack;
		nBack = temp;
	}

	poly.initialize(back,  nBack  + 2);
	poly.initialize(front, nFront + 2);


	offset = (rel0 == PLANAR)? split0 + 1 : split0;
	for (i = 0; i < nFront; i++){
		poly.copyVertex(front, i, offset + i);
	}

	offset = (rel1 == PLANAR)? split1 + 1 : split1;
	for (i = 0; i < nBack; i++){
		poly.copyVertex(back, i, (offset + i) % poly.nVertices);
	}

	j = (split0 == 0)? poly.nVertices - 1 : split0 - 1;

	k = getInterpolationConstant(poly.vertices[j], poly.vertices[split0]);
	poly.setInterpolatedVertex(front, nFront + 1, j, split0, k);
	poly.setInterpolatedVertex(back,  nBack,      j, split0, k);

	k = getInterpolationConstant(poly.vertices[split1-1], poly.vertices[split1]);
	poly.setInterpolatedVertex(front, nFront,    split1-1, split1, k);
	poly.setInterpolatedVertex(back,  nBack + 1, split1-1, split1, k);
	
	back->finalize();
	front->finalize();

	if (relprev0 == FRONT){
		Polygon *temp = back;
		back = front;
		front = temp;
	}
	

	return CUTS;
}


float Polygon::distance(const Vertex &v){
	//return (v - vertices[0]) * normal;
	return (v * normal + offset);
}

RELATION Polygon::vertexRelation(const Vertex &v){
	//float rel = (v - vertices[0]) * normal;
	float rel = (v * normal + offset);

	if (fabs(rel) < 0.03f) return PLANAR;
	if (rel > 0) return FRONT; else return BACK;
}

RELATION Polygon::polygonRelation(const Polygon &other){
	RELATION rel;
	int i, front, back;

	front = back = 0;

	for (i = 0; i < other.nVertices; i++){
		rel = vertexRelation(other.vertices[i]);
		if (rel == FRONT) front++;
		if (rel == BACK)  back++;
	}
	if (front > 0){
		if (back > 0) return CUTS; else return FRONT;
	} else {
		if (back > 0) return BACK; else return PLANAR;
	}
}


void Polygon::draw(PASS pass, Vertex &pos, TextureHandler *texHandler, Set <Light *> &lights){
	RELATION rel = vertexRelation(pos);

	glColor3f(1,1,1);

	if (flags & PF_UNLIT){
		GLfloat u,v;
		glColor3f(1,1,1);
		texHandler->setTextures(texID);
		texHandler->apply();
		glBegin(GL_TRIANGLE_FAN);
		for (int i = 0; i < nVertices; i++){
			u = (vertices[i] - origo) * s * scaleS;
			v = (vertices[i] - origo) * t * scaleT;
			glTexCoord2f(u,v);
			glVertex3fv((GLfloat *) (vertices + i));
		}
		glEnd();
	} else {
		int i, j, len = lights.getSize();
		int nDrawn = 0;
		float lightSize, d;
		Vertex lightPos;
		
		
		glDepthMask(GL_FALSE);
		if (flags & PF_TRANSLUCENT){
			texHandler->setBlending(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		}
		
		
		for (i = 0; i < len; i++){
			lightPos  = lights[i]->getPosition();
			lightSize = lights[i]->getSize();
			d = distance(lightPos);
			
			if (flags & PF_DOUBLESIDED){
				if (fabsf(d) > lightSize) continue;
			} else {
				if (d < -0.1f * lightSize || d > lightSize) continue;
			}
			
			if (!sphereIntersects(lightPos, lightSize)) continue;
			
			lightSize = 0.5f / lightSize;
			
			if (nDrawn == 0){
				if (flags & PF_SPECULAR){
					texHandler->setTextures(1, bumpID, 0);
					texHandler->setFragmentShader(0);
				} else {
					texHandler->setTextures(2, bumpID);
					texHandler->setTexEnvs(TEXENV_DOT3_NEXT, TEXENV_MODULATE_PRIMARY);
				}
				texHandler->apply();
			} else if (nDrawn == 1){
				texHandler->changeBlending(GL_ONE_MINUS_DST_COLOR, GL_ONE);
			}
			
			glColor4fv((GLfloat *) &lights[i]->getColor());
			
			glBegin(GL_TRIANGLE_FAN);
			for (j = 0; j < nVertices; j++){
				glMultiTexCoord3f(GL_TEXTURE0_ARB, (lightPos - vertices[j]) * s      * lightSize + 0.5f,
												   (lightPos - vertices[j]) * t      * lightSize + 0.5f,
												   (lightPos - vertices[j]) * normal * lightSize + 0.5f);
				glMultiTexCoord2f(GL_TEXTURE1_ARB, (vertices[j] - origo) * s * scaleS,
												   (vertices[j] - origo) * t * scaleT);
				if (flags & PF_SPECULAR){
					Vertex posDir = pos - vertices[j];
					posDir.fastNormalize();
					Vertex lightDir = lightPos - vertices[j];
					lightDir.fastNormalize();
					Vertex halfVector = posDir + lightDir;
					halfVector.fastNormalize();
					
					glMultiTexCoord3f(GL_TEXTURE2_ARB, halfVector * s, halfVector * t, halfVector * normal);
				} 
				glVertex3fv((GLfloat *) (vertices + j));
			}
			glEnd();
			
			nDrawn++;
		}
		
		glDepthMask(GL_TRUE);
		if (nDrawn == 0){
			texHandler->apply();
			
			glColor3f(0,0,0);
			glBegin(GL_TRIANGLE_FAN);
			for (int i = 0; i < nVertices; i++){
				glVertex3fv((GLfloat *) (vertices + i));
			}
			glEnd();
		} else {
			texHandler->setTextures(texID);
			texHandler->setBlending(GL_DST_COLOR, GL_SRC_COLOR);
			texHandler->apply();

			GLfloat u,v;
			glColor3f(1,1,1);
			glBegin(GL_TRIANGLE_FAN);
			for (int i = 0; i < nVertices; i++){
				u = (vertices[i] - origo) * s * scaleS;
				v = (vertices[i] - origo) * t * scaleT;
				glTexCoord2f(u,v);
				glVertex3fv((GLfloat *) (vertices + i));
			}
			glEnd();
		}
	}


/*
    texHandler->apply();

	// Draw normals
	Vertex vp, vo(0,0,0);
	for (int i = 0; i < nVertices; i++){
		vo += vertices[i] / float(nVertices);
	}
    vo += normal * 0.1f;

	glBegin(GL_LINES);
		glColor3f(1,0,0);
		glVertex3fv((GLfloat *) &vo);
		vp = vo + 25 * s;
		glVertex3fv((GLfloat *) &vp);
	glEnd();
	glBegin(GL_LINES);
		glColor3f(0,1,0);
		glVertex3fv((GLfloat *) &vo);
		vp = vo + 25 * t;
		glVertex3fv((GLfloat *) &vp);
	glEnd();
	glBegin(GL_LINES);
		glColor3f(1,1,1);
		glVertex3fv((GLfloat *) &vo);
		vp = vo + 25 * normal;
		glVertex3fv((GLfloat *) &vp);
	glEnd();
*/

}

void Polygon::drawFogDepthAlpha(Vertex &pos, float constDist){
	glBegin(GL_TRIANGLE_FAN);
	for (int i = 0; i < nVertices; i++){
		glColor4f(1,1,1, (dist(pos, vertices[i]) - constDist) / 300);
		glVertex3fv((GLfloat *) (vertices + i));
	}
	glEnd();
}

/* ########################################################################################################### */
