                  /*--.          .-"-. 
                 /   ._.        /  ` \
                 \_  (__\       \_v_/
                 //   \\        //   \\
                ((     ))      ((     ))
 --""---""----""---""--
                  |||            |||                                               
   Authors note:   |              |                                                
                                                                                   
   This file is part of work done by Emil Persson, aka Humus.                      
   You're free to use the code in any way you like, modified, unmodified           
   or cut'n'pasted into your own work. But for the good of the whole               
   programming community I will ask you to as long as possible and to the          
   extent you find suitable to follow these points:                                
    * If you use this file and it's contents unmodified, please leave this note.   
    * If you use a major part of the file in your app, please credit the author(s) 
      in one way or the other as you find suitable, for instance a small notice    
      in the readme file may do, or perhaps a link to the authors site.            
    * Share your work and ideas too as much as you can.                            
     _______                                                                       
   /` _____ `\;,    Humus                                                          
  (__(^===^)__)';,  emiper-8@student.luth.se                         ___           
    /  :::  \   ,;  ICQ #47010716                                  /^   ^\         
   |   :::   | ,;'                                                (     )        
 '._______.'`--oOo--(_)--oOo--*/


#ifndef _POLYGON_H_
#define _POLYGON_H_

enum RELATION {BACK = -1, PLANAR = 0, FRONT = 1, CUTS = 2};

#define PF_DOUBLESIDED 1
#define PF_UNLIT       2
#define PF_TRANSLUCENT 4
#define PF_DIFFUSE     8
#define PF_SPECULAR    16
#define PF_NONBLOCKING 32

/* ########################################################################################################### */

class Polygon {
protected:
	Vertex *vertices;
	int nVertices;

	Vertex normal;
	float offset;

	int flags;


	Vertex origo, s,t;
    float scaleS, scaleT;

	Vertex lightOrigo, lightS, lightT;

	TextureID texID;
    TextureID bumpID;

public:
	Polygon(){
		nVertices = 0;
		vertices = NULL;
		flags = 0;
	}

	Polygon(unsigned int nVerts){
		nVertices = nVerts;
		vertices = new Vertex[nVerts];
		flags = 0;
	}

	~Polygon(){
		if (vertices != NULL) delete vertices;
	}

	Polygon(Vertex v0, Vertex v1, Vertex v2){
		nVertices = 3;
		vertices = new Vertex[3];
		vertices[0] = v0;
		vertices[1] = v1;
		vertices[2] = v2;
		normal = cross(v1 - v0, v2 - v0);
		normal.normalize();
		offset = -(normal * vertices[0]);
		flags = 0;
	}

	Polygon(Vertex v0, Vertex v1, Vertex v2, Vertex v3){
		nVertices = 4;
		vertices = new Vertex[4];
		vertices[0] = v0;
		vertices[1] = v1;
		vertices[2] = v2;
		vertices[3] = v3;
		normal = cross(v1 - v0, v2 - v0);
		normal.normalize();
		offset = -(normal * vertices[0]);
		flags = 0;
	}

	void setFlags(int flag, bool value){
		if (value){
			flags |= flag;
		} else {
			flags &= ~flag;
		}
	}

	void setVertex(const int index, const Vertex &vertex){
		if (index == nVertices){
			nVertices++;
			vertices = (Vertex *) realloc(vertices, nVertices * sizeof(Vertex));
		}
		vertices[index] = vertex;
	}

	void setNormal(const Vertex &norm){
		normal = norm;
		offset = -(normal * vertices[0]);
	}

	Vertex getNormal(){
		return normal;
	}

    float getOffset(){
        return offset;
    }

    void setTextures(TextureID baseTexture, TextureID bumpMap){
	    texID  = baseTexture;
        bumpID = bumpMap;
    }

	void setTexCoordSystem(const Vertex &Origo, const Vertex &S, const Vertex &T){
		origo = Origo;
		s = S;
		t = T;
        scaleS = length(s);
        scaleT = length(t);
        s /= scaleS;
        t /= scaleT;
	}

	int getFlags(){
		return flags;
	}

    bool isInFrustum(Frustum &frustum){
        return frustum.polygonInFrustum(vertices, nVertices);
    }

	void initialize(Polygon *&poly, unsigned int nVerts);
	void finalize();
	void copyVertex(Polygon *&poly, unsigned int destIndex, unsigned int srcIndex);
	void setInterpolatedVertex(Polygon *&poly, unsigned int destIndex, int srcIndex0, int srcIndex1, float k);

	float getInterpolationConstant(Vertex &v0, Vertex &v1){
		return (normal * v0 + offset) / (normal * (v0 - v1));
	}

	bool linePassesThrough(const Vertex &v0, const Vertex &v1, Vertex &result);
    bool sphereIntersects(const Vertex &v, const float radius);

	RELATION split(Polygon &poly, Polygon *&back, Polygon *&front);

	float distance(const Vertex &v);
	RELATION vertexRelation(const Vertex &v);
	RELATION polygonRelation(const Polygon &other);

	void draw(PASS pass, Vertex &pos, TextureHandler *texHandler, Set <Light *> &lights);
	void drawFogDepthAlpha(Vertex &pos, float constDist);
};


#endif // _POLYGON_H_
